@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2017	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -21,7 +21,6 @@
 #include "pool.h"
 #include "pool_config.h"
 #include "pool_process_context.h"
-#include "pool_path.h"
 
 #include <ctype.h>
 #include <sys/types.h>
@@ -63,6 +62,7 @@
 #include "pool_passwd.h"
 #include "pool_memqcache.h"
 #include "watchdog/wd_ext.h"
+
 /*
  * Process pending signal actions.
  */
@@ -97,7 +97,6 @@
 #define PGPOOLMAXLITSENQUEUELENGTH 10000
 static void daemonize(void);
 static int read_pid_file(void);
-static char *get_pid_file_path(void);
 static void write_pid_file(void);
 static int read_status_file(bool discard_status);
 static int write_status_file(void);
@@ -115,7 +114,6 @@ static int pool_pause(struct timeval *timeout);
 static void kill_all_children(int sig);
 static int get_next_master_node(void);
 static pid_t fork_follow_child(int old_master, int new_primary, int old_primary);
-static void myunlink(const char* path);
 
 static RETSIGTYPE exit_handler(int sig);
 static RETSIGTYPE reap_handler(int sig);
@@ -141,7 +139,6 @@ static struct sockaddr_un un_addr;		/* unix domain socket path */
 static struct sockaddr_un pcp_un_addr;  /* unix domain socket path for PCP */
 
 ProcessInfo *process_info;	/* Per child info table on shmem */
-struct timeval random_start_time;
 
 /*
  * Private copy of backend status
@@ -162,16 +159,14 @@ static int follow_pid; /* pid for child process handling follow command */
 static int pcp_pid; /* pid for child process handling PCP */
 static int pcp_unix_fd; /* unix domain socket fd for PCP (not used) */
 static int pcp_inet_fd; /* inet domain socket fd for PCP */
-
-char *pcp_conf_file = NULL; /* absolute path of the pcp.conf */
-char *conf_file = NULL;		/* absolute path of the pgpool.conf */
-char *hba_file = NULL;		/* absolute path of the hba.conf */
-char *base_dir = NULL;		/* The working dir from where pgpool was invoked from */
+static char pcp_conf_file[POOLMAXPATHLEN+1]; /* path for pcp.conf */
+static char conf_file[POOLMAXPATHLEN+1];
+static char hba_file[POOLMAXPATHLEN+1];
 
 static int exiting = 0;		/* non 0 if I'm exiting */
 static int switching = 0;		/* non 0 if I'm fail overing or degenerating */
 
-static int clear_cache = 0;		/* non 0 if clear chache option (-c) is given */
+static int clear_cache = 0;		/* non 0 if clear cache option (-c) is given */
 static int not_detach = 0;		/* non 0 if non detach option (-n) is given */
 
 static int stop_sig = SIGTERM;	/* stopping signal default value */
@@ -190,7 +185,6 @@ int my_proc_id;
 static BackendStatusRecord backend_rec;	/* Backend status record */
 
 static pid_t worker_pid; /* pid of worker process */
-static pid_t watchdog_pid; /* pid of watchdog process */
 
 BACKEND_STATUS* my_backend_status[MAX_NUM_BACKENDS];		/* Backend status buffer */
 int my_master_node_id;		/* Master node id buffer */
@@ -214,10 +208,6 @@ int main(int argc, char **argv)
 	bool discard_status = false;
 	bool retrying;
 	bool clear_memcache_oidmaps = false;
-	char pcp_conf_file_path[POOLMAXPATHLEN+1];
-	char conf_file_path[POOLMAXPATHLEN+1];
-	char hba_file_path[POOLMAXPATHLEN+1];
-
 
 	static struct option long_options[] = {
 		{"hba-file", required_argument, NULL, 'a'},
@@ -233,13 +223,13 @@ int main(int argc, char **argv)
 		{"version", no_argument, NULL, 'v'},
 		{NULL, 0, NULL, 0}
 	};
-	
+
 	myargc = argc;
 	myargv = argv;
 
-	snprintf(conf_file_path, sizeof(conf_file_path), "%s/%s", DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
-	snprintf(pcp_conf_file_path, sizeof(pcp_conf_file_path), "%s/%s", DEFAULT_CONFIGDIR, PCP_PASSWD_FILE_NAME);
-	snprintf(hba_file_path, sizeof(hba_file_path), "%s/%s", DEFAULT_CONFIGDIR, HBA_CONF_FILE_NAME);
+	snprintf(conf_file, sizeof(conf_file), "%s/%s", DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
+	snprintf(pcp_conf_file, sizeof(pcp_conf_file), "%s/%s", DEFAULT_CONFIGDIR, PCP_PASSWD_FILE_NAME);
+	snprintf(hba_file, sizeof(hba_file), "%s/%s", DEFAULT_CONFIGDIR, HBA_CONF_FILE_NAME);
 
     while ((opt = getopt_long(argc, argv, "a:cdf:F:hm:nDCv", long_options, &optindex)) != -1)
 	{
@@ -251,7 +241,7 @@ int main(int argc, char **argv)
 					usage();
 					exit(1);
 				}
-				strlcpy(hba_file_path, optarg, sizeof(hba_file_path));
+				strlcpy(hba_file, optarg, sizeof(hba_file));
 				break;
 
 			case 'c':			/* clear cache option */
@@ -268,7 +258,7 @@ int main(int argc, char **argv)
 					usage();
 					exit(1);
 				}
-				strlcpy(conf_file_path, optarg, sizeof(conf_file_path));
+				strlcpy(conf_file, optarg, sizeof(conf_file));
 				break;
 
 			case 'F':   /* specify PCP password file */
@@ -277,7 +267,7 @@ int main(int argc, char **argv)
 					usage();
 					exit(1);
 				}
-				strlcpy(pcp_conf_file_path, optarg, sizeof(pcp_conf_file_path));
+				strlcpy(pcp_conf_file, optarg, sizeof(pcp_conf_file));
 				break;
 
 			case 'h':
@@ -326,38 +316,6 @@ int main(int argc, char **argv)
 		}
 	}
 
-	/* For PostmasterRandom */
-	gettimeofday(&random_start_time, NULL);
-
-	/* load the CWD before it is changed */
-	base_dir = get_current_working_dir();
-	if (base_dir == NULL)
-	{
-		pool_error("could not get current working directory. reason:%s\nExiting...",strerror(errno));
-		exit(1);
-	}
-	/* convert all the paths to absolute paths*/
-	conf_file = make_absolute_path(conf_file_path, base_dir);
-	if (conf_file == NULL)
-	{
-		pool_error("could not get the absolute file path for pgpool conf file. Exiting...");
-		exit(1);
-	}
-
-	pcp_conf_file = make_absolute_path(pcp_conf_file_path, base_dir);
-	if (pcp_conf_file == NULL)
-	{
-		pool_error("could not get the absolute file path for pcp conf file. Exiting...");
-		exit(1);
-	}
-
-	hba_file = make_absolute_path(hba_file_path, base_dir);
-	if (hba_file == NULL)
-	{
-		pool_error("could not get the absolute file path for pool_hba conf file. Exiting...");
-		exit(1);
-	}
-
 #ifdef USE_SSL
 	/* global ssl init */
 	SSL_library_init();
@@ -370,7 +328,7 @@ int main(int argc, char **argv)
 		exit(1);
 
 	/*
-	 * Init debug level with -d option value
+	 * Override debug level
 	 */
 	pool_config->debug_level = debug_level;
 
@@ -390,6 +348,15 @@ int main(int argc, char **argv)
 		pool_config->logsyslog = 1;
 	}
 
+	/*
+	 * Override debug level
+	 */
+	if (pool_config->debug_level == 0)
+		pool_config->debug_level = debug_level;
+
+	if (pool_config->enable_pool_hba)
+		load_hba(hba_file);
+
 	/*
 	 * If a non-switch argument remains, then it should be either "reload" or "stop".
 	 */
@@ -419,6 +386,7 @@ int main(int argc, char **argv)
 		if (!strcmp(argv[optind], "stop"))
 		{
 			stop_me();
+			unlink(pool_config->pid_file_name);
 			pool_shmem_exit(0);
 			exit(0);
 		}
@@ -455,18 +423,15 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	if (pool_config->enable_pool_hba)
-		load_hba(hba_file);
-
 	/* check effective user id for watchdog */
 	/* watchdog must be started under the privileged user */
 	if (pool_config->use_watchdog )
 	{
-		/* check sticky bit of network interface control commands */
-		if (wd_chk_sticky() == 1)
+		/* check setuid bit of network interface control commands */
+		if (wd_chk_setuid() == 1)
 		{
-			/* if_up, if_down and arping command have a sticky bit */
-			pool_log("watchdog might call network commands which using sticky bit.");
+			/* if_up, if_down and arping command have a setuid bit */
+			pool_log("watchdog might call network commands which using setuid bit.");
 		}
 	}
 
@@ -553,7 +518,7 @@ int main(int argc, char **argv)
 	con_info = pool_shared_memory_create(size);
 	if (con_info == NULL)
 	{
-		pool_error("failed to allocate connection informations");
+		pool_error("failed to allocate connection information");
 		myexit(1);
 	}
 	memset(con_info, 0, size);
@@ -613,7 +578,7 @@ int main(int argc, char **argv)
 		if (pool_is_shmem_cache())
 		{
 			size_t size;
-			
+
 			size = pool_shared_memory_cache_size();
 			if (size == 0)
 			{
@@ -668,8 +633,7 @@ int main(int argc, char **argv)
 	/* start watchdog */
 	if (pool_config->use_watchdog )
 	{
-		watchdog_pid = wd_main(1);
-		if (watchdog_pid == 0)
+		if (!wd_main(1))
 		{
 			pool_error("wd_main error");
 			myexit(1);
@@ -789,7 +753,7 @@ int main(int argc, char **argv)
 					{
 						if (POOL_DISALLOW_TO_FAILOVER(BACKEND_INFO(sts).flag))
 						{
-							pool_log("health_check: %d failover is canceld because failover is disallowed", sts);
+							pool_log("health_check: %d failover is canceled because failover is disallowed", sts);
 						}
 						else if (retrycnt <= pool_config->health_check_max_retries)
 						{
@@ -869,7 +833,7 @@ int main(int argc, char **argv)
 
 			if (pool_config->health_check_timeout > 0)
 			{
-				/* seems ok. cancel health check timer */
+				/* seems OK. cancel health check timer */
 				pool_signal(SIGALRM, SIG_IGN);
 				CLEAR_ALARM;
 			}
@@ -975,23 +939,18 @@ static void daemonize(void)
 	write_pid_file();
 	rc_chdir = chdir("/");
 
+	/* redirect stdin, stdout and stderr to /dev/null */
 	i = open("/dev/null", O_RDWR);
-	if(i < 0)
-	{
-		pool_error("Failed to open \"/dev/null\" reason:%s", strerror(errno));
-	}
-	else
-	{
-		dup2(i, 0);
-		dup2(i, 1);
-		dup2(i, 2);
-		close(i);
-	}
+	dup2(i, 0);
+	dup2(i, 1);
+	dup2(i, 2);
+
 	/* close syslog connection for daemonizing */
 	if (pool_config->logsyslog) {
 		closelog();
 	}
 
+	/* close other file descriptors */
     fdlimit = sysconf(_SC_OPEN_MAX);
     for (i = 3; i < fdlimit; i++)
 		close(i);
@@ -1010,7 +969,6 @@ static void daemonize(void)
 static void stop_me(void)
 {
 	pid_t pid;
-	char *pid_file;
 
 	pid = read_pid_file();
 	if (pid < 0)
@@ -1035,58 +993,6 @@ static void stop_me(void)
 		sleep(1);
 	}
 	fprintf(stderr, "done.\n");
-	pid_file = get_pid_file_path();
-	if(pid_file)
-	{
-		unlink(pid_file);
-		free(pid_file);
-	}
-	else
-	{
-		pool_error("failed to get pid file path from %s",
-				   pool_config->pid_file_name);
-	}
-}
-
-/*
- * The function returns the malloc'd copy of pid_file_path,
- * caller must free it after use
- */
-static char *get_pid_file_path(void)
-{
-	char *new = NULL;
-	if (!is_absolute_path(pool_config->pid_file_name))
-	{
-		/*
-		 * some implementations of dirname() may modify the
-		 * string argument passed to it, so do not use the original
-		 * conf_file as an argument
-		 */
-		char *conf_file_copy = strdup(conf_file);
-		char *conf_dir = dirname(conf_file_copy);
-		size_t  path_size;
-		if (conf_dir == NULL)
-		{
-			pool_error("could not read pid file as %s. reason: %s",
-					   pool_config->pid_file_name, strerror(errno));
-			return NULL;
-		}
-		path_size = strlen(conf_dir) + strlen(pool_config->pid_file_name) + 1 + 1;
-		new = malloc(path_size);
-		if (new == NULL)
-		{
-			pool_error("out of memory");
-			exit(1);
-		}
-		snprintf(new, path_size, "%s/%s",conf_dir, pool_config->pid_file_name);
-
-		pool_debug("pid file location is \"%s\"",new);
-	}
-	else
-	{
-		new = strdup(pool_config->pid_file_name);
-	}
-	return new;
 }
 
 /*
@@ -1097,38 +1003,26 @@ static int read_pid_file(void)
 	int fd;
 	int readlen;
 	char pidbuf[128];
-	char *pid_file = get_pid_file_path();
 
-	if (pid_file == NULL)
-	{
-		pool_error("failed to get pid file path from %s",
-				   pool_config->pid_file_name);
-		return -1;
-	}
-
-	fd = open(pid_file, O_RDONLY);
+	fd = open(pool_config->pid_file_name, O_RDONLY);
 	if (fd == -1)
 	{
-		free(pid_file);
 		return -1;
 	}
 	if ((readlen = read(fd, pidbuf, sizeof(pidbuf))) == -1)
 	{
 		pool_error("could not read pid file as %s. reason: %s",
-				   pid_file, strerror(errno));
-		free(pid_file);
+				   pool_config->pid_file_name, strerror(errno));
 		close(fd);
 		return -1;
 	}
 	else if (readlen == 0)
 	{
 		pool_error("EOF detected while reading pid file as %s. reason: %s",
-				   pid_file, strerror(errno));
-		free(pid_file);
+				   pool_config->pid_file_name, strerror(errno));
 		close(fd);
 		return -1;
 	}
-	free(pid_file);
 	close(fd);
 	return(atoi(pidbuf));
 }
@@ -1140,21 +1034,12 @@ static void write_pid_file(void)
 {
 	int fd;
 	char pidbuf[128];
-	char *pid_file = get_pid_file_path();
 
-	if (pid_file == NULL)
-	{
-		pool_error("failed to get pid file path from %s",
-				   pool_config->pid_file_name);
-		exit(1);
-	}
-
-	fd = open(pid_file, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
+	fd = open(pool_config->pid_file_name, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
 	if (fd == -1)
 	{
 		pool_error("could not open pid file as %s. reason: %s",
-				   pid_file, strerror(errno));
-		free(pid_file);
+				   pool_config->pid_file_name, strerror(errno));
 		pool_shmem_exit(1);
 		exit(1);
 	}
@@ -1162,8 +1047,7 @@ static void write_pid_file(void)
 	if (write(fd, pidbuf, strlen(pidbuf)+1) == -1)
 	{
 		pool_error("could not write pid file as %s. reason: %s",
-				   pid_file, strerror(errno));
-		free(pid_file);
+				   pool_config->pid_file_name, strerror(errno));
 		close(fd);
 		pool_shmem_exit(1);
 		exit(1);
@@ -1171,23 +1055,20 @@ static void write_pid_file(void)
 	if (fsync(fd) == -1)
 	{
 		pool_error("could not fsync pid file as %s. reason: %s",
-				   pid_file, strerror(errno));
-		free(pid_file);
+				   pool_config->pid_file_name, strerror(errno));
 		close(fd);
 		pool_shmem_exit(1);
 		exit(1);
-	}	
+	}
 	if (close(fd) == -1)
 	{
 		pool_error("could not close pid file as %s. reason: %s",
-				   pid_file, strerror(errno));
+				   pool_config->pid_file_name, strerror(errno));
 		pool_shmem_exit(1);
 		exit(1);
 	}
-	free(pid_file);
 }
 
-
 /*
 * Read the status file
 */
@@ -1267,7 +1148,6 @@ static int read_status_file(bool discard_status)
 static int write_status_file(void)
 {
 	FILE *fd;
-	int fdnum;
 	char fnamebuf[POOLMAXPATHLEN];
 	int i;
 
@@ -1293,32 +1173,6 @@ static int write_status_file(void)
 		fclose(fd);
 		return -1;
 	}
-
-	if (fflush(fd) != 0)
-	{
-		pool_error("Could not flush backend status file as %s. reason: %s",
-				   fnamebuf, strerror(errno));
-		fclose(fd);
-		return -1;
-	}
-
-	fdnum = fileno(fd);
-	if (fdnum < 0)
-	{
-		pool_error("Failed to get file number of %s. reason: %s",
-				   fnamebuf, strerror(errno));
-		fclose(fd);
-		return -1;
-	}
-
-	if (fsync(fdnum) != 0)
-	{
-		pool_error("Failed to fsync() %s. reason: %s",
-				   fnamebuf, strerror(errno));
-		fclose(fd);
-		return -1;
-	}
-
 	fclose(fd);
 	return 0;
 }
@@ -1556,14 +1410,12 @@ static int create_unix_domain_socket(struct sockaddr_un un_addr_tmp)
 static void myunlink(const char* path)
 {
 	if (unlink(path) == 0) return;
-		pool_error("unlink(%s) failed: %s", path, strerror(errno));
+	pool_error("unlink(%s) failed: %s", path, strerror(errno));
 }
 
 static void myexit(int code)
 {
 	int i;
-	pid_t wpid;
-	char *pid_file;
 
 	if (getpid() != mypid)
 		return;
@@ -1581,33 +1433,16 @@ static void myexit(int code)
 		}
 
 		/* wait for all children to exit */
-		do
-		{
-			pid_t ret_pid;
-
-			wpid = waitpid(-1, &ret_pid, 0);
-		} while (wpid > 0 || (wpid == -1 && errno == EINTR));
-
-		if (wpid == -1 && errno != ECHILD)
+		while (wait(NULL) > 0)
+			;
+		if (errno != ECHILD)
 			pool_error("wait() failed. reason:%s", strerror(errno));
-
 		POOL_SETMASK(&UnBlockSig);
 	}
 
 	myunlink(un_addr.sun_path);
 	myunlink(pcp_un_addr.sun_path);
-
-	pid_file = get_pid_file_path();
-	if(pid_file)
-	{
-		unlink(pid_file);
-		free(pid_file);
-	}
-	else
-	{
-		pool_error("failed to get pid file path from %s",
-				   pool_config->pid_file_name);
-	}
+	myunlink(pool_config->pid_file_name);
 
 	write_status_file();
 
@@ -1629,25 +1464,11 @@ void notice_backend_error(int node_id)
 	}
 }
 
-/*
- * degenerate_backend_set_ex:
- *
- * The function signals/verifies the node down request.
- * The request is then processed by failover function.
- *
- * node_id_set: array of node ids to be registered for NODE DOWN operation
- * count:       number of elements in node_id_set array
- * test_only:   When set, function only checks if NODE DOWN operation can be
- *              executed on provided node ids and never registers the operation
- *              request.
- *              For test_only case function returs false as
- *              soon as first non complient node in node_id_set is found
- */
-bool degenerate_backend_set_ex(int *node_id_set, int count, bool test_only)
+/* notice backend connection error using SIGUSR1 */
+void degenerate_backend_set(int *node_id_set, int count)
 {
 	pid_t parent = getppid();
 	int i;
-	bool ret = false;
 	bool need_signal = false;
 #ifdef HAVE_SIGPROCMASK
 	sigset_t oldmask;
@@ -1657,68 +1478,47 @@ bool degenerate_backend_set_ex(int *node_id_set, int count, bool test_only)
 
 	if (pool_config->parallel_mode)
 	{
-		return false;
-	}
-	if (test_only == false)
-	{
-		POOL_SETMASK2(&BlockSig, &oldmask);
-		pool_semaphore_lock(REQUEST_INFO_SEM);
-		Req_info->kind = NODE_DOWN_REQUEST;
+		return;
 	}
 
+	POOL_SETMASK2(&BlockSig, &oldmask);
+	pool_semaphore_lock(REQUEST_INFO_SEM);
+	Req_info->kind = NODE_DOWN_REQUEST;
 	for (i = 0; i < count; i++)
 	{
 		if (node_id_set[i] < 0 || node_id_set[i] >= MAX_NUM_BACKENDS ||
 			!VALID_BACKEND(node_id_set[i]))
 		{
 			pool_log("degenerate_backend_set: node %d is not valid backend.", i);
-			if (test_only)
-				return false;
 			continue;
 		}
 
 		if (POOL_DISALLOW_TO_FAILOVER(BACKEND_INFO(node_id_set[i]).flag))
 		{
-			pool_log("degenerate_backend_set: %d failover request from pid %d is canceld because failover is disallowed", node_id_set[i], getpid());
-			if (test_only)
-				return false;
+			pool_log("degenerate_backend_set: %d failover request from pid %d is canceled because failover is disallowed", node_id_set[i], getpid());
 			continue;
 		}
 
-		if (test_only == false)
-		{
-			pool_log("degenerate_backend_set: %d fail over request from pid %d", node_id_set[i], getpid());
-			Req_info->node_id[i] = node_id_set[i];
-		}
+		pool_log("degenerate_backend_set: %d fail over request from pid %d", node_id_set[i], getpid());
+		Req_info->node_id[i] = node_id_set[i];
 		need_signal = true;
 	}
 
 	if (need_signal)
 	{
-		if (test_only)
-			return true;
 		if (!pool_config->use_watchdog || WD_OK == wd_degenerate_backend_set(node_id_set, count))
 		{
 			kill(parent, SIGUSR1);
-			ret = true;
+		}
+		else
+		{
+			pool_log("degenerate_backend_set: failover request from pid %d is canceled by other pgpool", getpid());
+			memset(Req_info->node_id, -1, sizeof(int) * MAX_NUM_BACKENDS);
 		}
 	}
-	else if (test_only)
-		return false;
 
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 	POOL_SETMASK(&oldmask);
-
-	return ret;
-}
-
-/*
- * wrapper over degenerate_backend_set_ex function to signal
- * NODE down operation request
- */
-void degenerate_backend_set(int *node_id_set, int count)
-{
-	degenerate_backend_set_ex(node_id_set, count, false);
 }
 
 /* send promote node request using SIGUSR1 */
@@ -1746,6 +1546,12 @@ void promote_backend(int node_id)
 	{
 		kill(parent, SIGUSR1);
 	}
+	else
+	{
+		pool_log("promote_backend: promote request from pid %d is canceled by other pgpool", getpid());
+		Req_info->node_id[0] = -1;
+	}
+
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 }
 
@@ -1758,15 +1564,18 @@ void send_failback_request(int node_id)
 	Req_info->kind = NODE_UP_REQUEST;
 	Req_info->node_id[0] = node_id;
 
-    if (node_id < 0 || node_id >= MAX_NUM_BACKENDS || 
+    if (node_id < 0 || node_id >= MAX_NUM_BACKENDS ||
 		(RAW_MODE && BACKEND_INFO(node_id).backend_status != CON_DOWN && VALID_BACKEND(node_id)))
 	{
 		pool_error("send_failback_request: node %d is alive.", node_id);
+		Req_info->node_id[0] = -1;
 		return;
 	}
 
 	if (pool_config->use_watchdog && WD_OK != wd_send_failback_request(node_id))
 	{
+		pool_log("send_failback_request: failback request from pid %d is canceled by other pgpool", getpid());
+		Req_info->node_id[0] = -1;
 		return;
 	}
 	kill(parent, SIGUSR1);
@@ -1775,7 +1584,6 @@ void send_failback_request(int node_id)
 static RETSIGTYPE exit_handler(int sig)
 {
 	int i;
-	pid_t wpid;
 
 	POOL_SETMASK(&AuthBlockSig);
 
@@ -1787,6 +1595,7 @@ static RETSIGTYPE exit_handler(int sig)
 	{
 		pool_debug("exit_handler: I am not parent");
 		POOL_SETMASK(&UnBlockSig);
+		pool_shmem_exit(0);
 		exit(0);
 	}
 
@@ -1805,11 +1614,6 @@ static RETSIGTYPE exit_handler(int sig)
 
 	exiting = 1;
 
-	/* Close listen socket */
-	pool_log("pgpool main: close listen socket");
-	close(inet_fd);
-	close(unix_fd);
-
 	for (i = 0; i < pool_config->num_init_children; i++)
 	{
 		pid_t pid = process_info[i].pid;
@@ -1824,20 +1628,15 @@ static RETSIGTYPE exit_handler(int sig)
 
 	if (pool_config->use_watchdog)
 	{
-		pool_log("watchdog_pid: %d", watchdog_pid);
-		kill(watchdog_pid, sig);
+		wd_kill_watchdog(sig);
 	}
 
 	POOL_SETMASK(&UnBlockSig);
 
-    do
-    {
-		pid_t ret_pid;
+	while (wait(NULL) > 0)
+		;
 
-        wpid = waitpid(-1, &ret_pid, 0);
-    } while (wpid > 0 || (wpid == -1 && errno == EINTR));
-
-    if (wpid == -1 && errno != ECHILD)
+	if (errno != ECHILD)
 		pool_error("wait() failed. reason:%s", strerror(errno));
 
 	process_info = NULL;
@@ -1898,6 +1697,7 @@ static void failover(void)
 {
 	int i;
 	int node_id;
+	bool by_health_check;
 	int new_master;
 	int new_primary;
 	int nodes[MAX_NUM_BACKENDS];
@@ -1959,34 +1759,32 @@ static void failover(void)
 	Req_info->switching = true;
 	node_id = Req_info->node_id[0];
 
-	/* Perform failover with health check alarm
-	 * disabled
-	 */
-	if (pool_config->health_check_timeout > 0)
+	/* start of command inter-lock with watchdog */
+	if (pool_config->use_watchdog)
 	{
-		pool_signal(SIGALRM, SIG_IGN);
-		CLEAR_ALARM;
-		health_check_timer_expired = 0;
+		by_health_check = (!failover_request && Req_info->kind==NODE_DOWN_REQUEST);
+		wd_start_interlock(by_health_check);
 	}
 
 	/* failback request? */
 	if (Req_info->kind == NODE_UP_REQUEST)
 	{
-		if (node_id < 0 || node_id >= MAX_NUM_BACKENDS ||
+		if (node_id >= MAX_NUM_BACKENDS ||
 			(Req_info->kind == NODE_UP_REQUEST && !(RAW_MODE &&
             BACKEND_INFO(node_id).backend_status == CON_DOWN) && VALID_BACKEND(node_id)) ||
 			(Req_info->kind == NODE_DOWN_REQUEST && !VALID_BACKEND(node_id)))
 		{
 			pool_semaphore_unlock(REQUEST_INFO_SEM);
-
-			if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
-				pool_error("failover_handler: invalid node_id %d MAX_NUM_BACKENDS: %d", node_id, MAX_NUM_BACKENDS);
-			else
-				pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
-						   BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
+			pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
+					   BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
 			kill(pcp_pid, SIGUSR2);
 			switching = 0;
 			Req_info->switching = false;
+
+			/* end of command inter-lock */
+			if (pool_config->use_watchdog)
+				wd_leave_interlock();
+
 			return;
 		}
 
@@ -1994,8 +1792,22 @@ static void failover(void)
 				 BACKEND_INFO(node_id).backend_hostname,
 				 BACKEND_INFO(node_id).backend_port);
 		BACKEND_INFO(node_id).backend_status = CON_CONNECT_WAIT;	/* unset down status */
-		trigger_failover_command(node_id, pool_config->failback_command,
-								 MASTER_NODE_ID, get_next_master_node(), PRIMARY_NODE_ID);
+
+		/* wait for failback command lock or to be lock holder */
+		if (pool_config->use_watchdog && !wd_am_I_lock_holder())
+		{
+			wd_wait_for_lock(WD_FAILBACK_COMMAND_LOCK);
+		}
+		/* execute failback command if lock holder */
+		if (!pool_config->use_watchdog || wd_am_I_lock_holder())
+		{
+			trigger_failover_command(node_id, pool_config->failback_command,
+								 	MASTER_NODE_ID, get_next_master_node(), PRIMARY_NODE_ID);
+
+			/* unlock failback command */
+			if (pool_config->use_watchdog)
+				wd_unlock(WD_FAILBACK_COMMAND_LOCK);
+		}
 	}
 	else if (Req_info->kind == PROMOTE_NODE_REQUEST)
 	{
@@ -2012,6 +1824,11 @@ static void failover(void)
 			kill(pcp_pid, SIGUSR2);
 			switching = 0;
 			Req_info->switching = false;
+
+			/* end of command inter-lock */
+			if (pool_config->use_watchdog)
+				wd_leave_interlock();
+
 			return;
 		}
 	}
@@ -2043,6 +1860,11 @@ static void failover(void)
 			kill(pcp_pid, SIGUSR2);
 			switching = 0;
 			Req_info->switching = false;
+
+			/* end of command inter-lock */
+			if (pool_config->use_watchdog)
+				wd_leave_interlock();
+
 			return;
 		}
 	}
@@ -2107,13 +1929,13 @@ static void failover(void)
    /* On 2011/5/2 Tatsuo Ishii says: if mode is streaming replication
 	* and request is NODE_UP_REQUEST(failback case) we don't need to
 	* restart all children. Existing session will not use newly
-	* attached node, but load balanced node is not changed util this
+	* attached node, but load balanced node is not changed until this
 	* session ends, so it's harmless anyway.
 	*/
 	if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP)	&&
 		Req_info->kind == NODE_UP_REQUEST)
 	{
-		pool_log("Do not restart children because we are failbacking node id %d host%s port:%d and we are in streaming replication mode", node_id, 
+		pool_log("Do not restart children because we are failbacking node id %d host%s port:%d and we are in streaming replication mode", node_id,
 				 BACKEND_INFO(node_id).backend_hostname,
 				 BACKEND_INFO(node_id).backend_port);
 
@@ -2137,14 +1959,29 @@ static void failover(void)
 		need_to_restart_children = true;
 	}
 
-	/* Exec failover_command if needed */
-	for (i = 0; i < pool_config->backend_desc->num_backends; i++)
+	/* wait for failover command lock or to be lock holder*/
+	if (pool_config->use_watchdog && !wd_am_I_lock_holder())
 	{
-		if (nodes[i])
-			trigger_failover_command(i, pool_config->failover_command,
-									 MASTER_NODE_ID, new_master, PRIMARY_NODE_ID);
+		wd_wait_for_lock(WD_FAILOVER_COMMAND_LOCK);
 	}
 
+	/* execute failover command if lock holder */
+	if (!pool_config->use_watchdog || wd_am_I_lock_holder())
+	{
+		/* Exec failover_command if needed */
+		for (i = 0; i < pool_config->backend_desc->num_backends; i++)
+		{
+			if (nodes[i])
+				trigger_failover_command(i, pool_config->failover_command,
+									 		MASTER_NODE_ID, new_master, PRIMARY_NODE_ID);
+		}
+
+		/* unlock failover command */
+		if (pool_config->use_watchdog)
+			wd_unlock(WD_FAILOVER_COMMAND_LOCK);
+	}
+
+
 /* no need to wait since it will be done in reap_handler */
 #ifdef NOT_USED
 	while (wait(NULL) > 0)
@@ -2156,25 +1993,10 @@ static void failover(void)
 
 	if (Req_info->kind == PROMOTE_NODE_REQUEST && VALID_BACKEND(node_id))
 		new_primary = node_id;
-
-	/*
-	 * If the down node was a standby node in streaming replication
-	 * mode, we can avoid calling find_primary_node_repeatedly() and
-	 * recognize the former primary as the new primary node, which
-	 * will reduce the time to process standby down.
-	 */
-	else if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) &&
-			 Req_info->kind == NODE_DOWN_REQUEST)
-	{
-		if (Req_info->primary_node_id != node_id)
-			new_primary = Req_info->primary_node_id;
-		else
-			new_primary =  find_primary_node_repeatedly();
-	}
 	else
 		new_primary =  find_primary_node_repeatedly();
 
-	/* 
+	/*
 	 * If follow_master_command is provided and in master/slave
 	 * streaming replication mode, we start degenerating all backends
 	 * as they are not replicated anymore.
@@ -2222,13 +2044,30 @@ static void failover(void)
 	memset(Req_info->node_id, -1, sizeof(int) * MAX_NUM_BACKENDS);
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 
-	/* exec follow_master_command */
-	if ((follow_cnt > 0) && (*pool_config->follow_master_command != '\0'))
+	/* wait for follow_master_command lock or to be lock holder */
+	if (pool_config->use_watchdog && !wd_am_I_lock_holder())
 	{
-		follow_pid = fork_follow_child(Req_info->master_node_id, new_primary,
-									   Req_info->primary_node_id);
+		wd_wait_for_lock(WD_FOLLOW_MASTER_COMMAND_LOCK);
 	}
 
+	/* execute follow_master_command */
+	if (!pool_config->use_watchdog || wd_am_I_lock_holder())
+	{
+		if ((follow_cnt > 0) && (*pool_config->follow_master_command != '\0'))
+		{
+			follow_pid = fork_follow_child(Req_info->master_node_id, new_primary,
+									   	Req_info->primary_node_id);
+		}
+
+		/* unlock follow_master_command  */
+		if (pool_config->use_watchdog)
+			wd_unlock(WD_FOLLOW_MASTER_COMMAND_LOCK);
+	}
+
+	/* end of command inter-lock */
+	if (pool_config->use_watchdog)
+		wd_end_interlock();
+
 	/* Save primary node id */
 	Req_info->primary_node_id = new_primary;
 	pool_log("failover: set new primary node: %d", Req_info->primary_node_id);
@@ -2245,6 +2084,19 @@ static void failover(void)
 	{
 		for (i=0;i<pool_config->num_init_children;i++)
 		{
+
+			/*
+			 * Try to kill pgpool child because previous kill signal
+			 * may not be received by pgpool child. This could happen
+			 * if multiple PostgreSQL are going down (or even starting
+			 * pgpool, without starting PostgreSQL can trigger this).
+			 * Child calls degenerate_backend() and it tries to aquire
+			 * semaphore to write a failover request. In this case the
+			 * signal mask is set as well, thus signals are never
+			 * received.
+			 */
+			kill(process_info[i].pid, SIGQUIT);
+
 			process_info[i].pid = fork_a_child(unix_fd, inet_fd, i);
 			process_info[i].start_time = time(NULL);
 		}
@@ -2309,7 +2161,7 @@ static void failover(void)
 				continue;
 			else
 			{
-				pool_error("failover: waitpid failed. reson: %s", strerror(errno));
+				pool_error("failover: waitpid failed. reason: %s", strerror(errno));
 				return;
 			}
 		}
@@ -2336,7 +2188,7 @@ static RETSIGTYPE health_check_timer_handler(int sig)
 
 /*
  * Check if we can connect to the backend
- * returns 0 for ok. otherwise returns backend id + 1
+ * returns 0 for OK. otherwise returns backend id + 1
  */
 static int health_check(void)
 {
@@ -2379,7 +2231,7 @@ static int health_check(void)
 			bkinfo->backend_status == CON_DOWN)
 			continue;
 
-		slot = make_persistent_db_connection(i, bkinfo->backend_hostname,
+		slot = make_persistent_db_connection(bkinfo->backend_hostname,
 											 bkinfo->backend_port,
 											 dbname,
 											 pool_config->health_check_user,
@@ -2418,7 +2270,7 @@ static int health_check(void)
 
 /*
  * check if we can connect to the SystemDB
- * returns 0 for ok. otherwise returns -1
+ * returns 0 for OK. otherwise returns -1
  */
 static int
 system_db_health_check(void)
@@ -2537,7 +2389,7 @@ static void reaper(void)
 			/* Child terminated by segmentation fault. Report it */
 			pool_error("Child process %d was terminated by segmentation fault", pid);
 		}
-			
+
 		/* if exiting child process was PCP handler */
 		if (pid == pcp_pid)
 		{
@@ -2563,6 +2415,17 @@ static void reaper(void)
 
 			pool_log("fork a new worker child pid %d", worker_pid);
 		}
+
+		/* exiting process was watchdog process */
+		else if (pool_config->use_watchdog && wd_is_watchdog_pid(pid))
+		{
+			if (!wd_reaper_watchdog(pid, status))
+			{
+				pool_error("wd_reaper failed");
+				myexit(1);
+			}
+		}
+
 		else
 		{
 			if (WIFSIGNALED(status))
@@ -2726,7 +2589,7 @@ static void kill_all_children(int sig)
  * recovery processing) or config file reload request(SIGHUP) has been
  * occurred.  In this case this function returns 1.
  * otherwise 0: (no signal event occurred), -1: (error)
- * XXX: is it ok that select(2) error is ignored here?
+ * XXX: is it OK that select(2) error is ignored here?
  */
 static int pool_pause(struct timeval *timeout)
 {
@@ -2816,7 +2679,7 @@ static int trigger_failover_command(int node, const char *command_line,
 		return 0;
 
 	/* check failed nodeID */
-	if (node < 0 || node >= NUM_BACKENDS)
+	if (node < 0 || node > NUM_BACKENDS)
 		return -1;
 
 	info = pool_get_node_info(node);
@@ -2864,8 +2727,8 @@ static int trigger_failover_command(int node, const char *command_line,
 						if (newmaster)
 							string_append_char(exec_cmd, newmaster->backend_hostname);
 						else
-							/* no vaid new master */
-							string_append_char(exec_cmd, "\"\"");
+							/* no valid new master */
+							string_append_char(exec_cmd, "");
 						break;
 
 					case 'm': /* new master node id */
@@ -2881,8 +2744,8 @@ static int trigger_failover_command(int node, const char *command_line,
 							string_append_char(exec_cmd, port_buf);
 						}
 						else
-							/* no vaid new master */
-							string_append_char(exec_cmd, "\"\"");
+							/* no valid new master */
+							string_append_char(exec_cmd, "");
 						break;
 
 					case 'R': /* new master database directory */
@@ -2890,8 +2753,8 @@ static int trigger_failover_command(int node, const char *command_line,
 						if (newmaster)
 							string_append_char(exec_cmd, newmaster->backend_data_directory);
 						else
-							/* no vaid new master */
-							string_append_char(exec_cmd, "\"\"");
+							/* no valid new master */
+							string_append_char(exec_cmd, "");
 						break;
 
 					case 'M': /* old master node id */
@@ -2940,7 +2803,8 @@ static int find_primary_node(void)
 {
 	BackendInfo *bkinfo;
 	POOL_CONNECTION_POOL_SLOT *s;
-	POOL_CONNECTION *con; 
+	POOL_CONNECTION *con;
+	POOL_STATUS status;
 	POOL_SELECT_RESULT *res;
 	bool is_standby;
 	int i;
@@ -2967,7 +2831,7 @@ static int find_primary_node(void)
 		is_standby = false;
 
 		bkinfo = pool_get_node_info(i);
-		s = make_persistent_db_connection(i, bkinfo->backend_hostname,
+		s = make_persistent_db_connection(bkinfo->backend_hostname,
 										  bkinfo->backend_port,
 										  "postgres",
 										  pool_config->sr_check_user,
@@ -2975,18 +2839,10 @@ static int find_primary_node(void)
 		if (!s)
 		{
 			pool_error("find_primary_node: make_persistent_connection failed");
-
-			/*
-			 * It is possible that a node is down even if
-			 * VALID_BACKEND tells it's valid.  This could happen
-			 * before health checking detects the failure.
-			 * Thus we should continue to look for primary node.
-			 */
-			continue;
+			return -1;
 		}
 		con = s->con;
-#ifdef NOT_USED
-		status = do_query(con, "SELECT count(*) FROM pg_catalog.pg_proc AS p WHERE p.proname = 'pgpool_walrecrunning'",
+		status = do_query(con, "SELECT pg_is_in_recovery()",
 						  &res, PROTO_MAJOR_V3);
 		if (res->numrows <= 0)
 		{
@@ -3000,39 +2856,11 @@ static int find_primary_node(void)
 		{
 			pool_log("find_primary_node: do_query returns NULL");
 		}
-		if (res->data[0] && !strcmp(res->data[0], "0"))
+		if (res->data[0] && !strcmp(res->data[0], "t"))
 		{
-			pool_log("find_primary_node: pgpool_walrecrunning does not exist");
-			free_select_result(res);
-			discard_persistent_db_connection(s);
-			return -1;
+			is_standby = true;
 		}
-#endif
-		if(do_query(con, "SELECT pg_is_in_recovery()", &res, PROTO_MAJOR_V3) == POOL_CONTINUE)
-		{
-			if (res->numrows <= 0)
-			{
-				pool_log("find_primary_node: do_query returns no rows");
-			}
-			if (res->data[0] == NULL)
-			{
-				pool_log("find_primary_node: do_query returns no data");
-			}
-			if (res->nullflags[0] == -1)
-			{
-				pool_log("find_primary_node: do_query returns NULL");
-			}
-			if (res->data[0] && !strcmp(res->data[0], "t"))
-			{
-				is_standby = true;
-			}
-		}
-		else
-		{
-			pool_log("find_primary_node: do_query failed");
-		}
-		if(res)
-			free_select_result(res);
+		free_select_result(res);
 		discard_persistent_db_connection(s);
 
 		/*
@@ -3074,8 +2902,15 @@ static int find_primary_node_repeatedly(void)
 		return -1;
 	}
 
+	/*
+	 * Try to find the new primary node and keep trying for
+	 * search_primary_node_timeout seconds.
+	 * search_primary_node_timeout = 0 means never timeout and keep searching
+	 * indefinitely
+	 */
 	pool_log("find_primary_node_repeatedly: waiting for finding a primary node");
-	for (sec = 0; sec < pool_config->recovery_timeout; sec++)
+	for (sec = 0; (pool_config->search_primary_node_timeout == 0 ||
+				sec < pool_config->search_primary_node_timeout); sec++)
 	{
 		node_id = find_primary_node();
 		if (node_id != -1)
