@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2016	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -18,7 +18,7 @@
  * suitability of this software for any purpose.  It is provided "as
  * is" without express or implied warranty.
  *
- * pool_auth.c: authenticaton stuff
+ * pool_auth.c: authentication stuff
  *
 */
 
@@ -52,8 +52,6 @@ static int read_password_packet(POOL_CONNECTION *frontend, int protoMajor, 	char
 static int send_password_packet(POOL_CONNECTION *backend, int protoMajor, char *password);
 static int send_auth_ok(POOL_CONNECTION *frontend, int protoMajor);
 
-static long PostmasterRandom(void);
-
 /*
  * After sending the start up packet to the backend, do the
  * authentication against backend. if success return 0 otherwise non
@@ -70,7 +68,7 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	int i;
 	StartupPacket *sp;
 
-	protoMajor = MASTER_CONNECTION(cp)->sp->major;
+	protoMajor = MAJOR(cp);
 
 	kind = pool_read_kind(cp);
 	if (kind < 0)
@@ -332,6 +330,12 @@ from pool_read_message_length and recheck the pg_hba.conf settings.");
 	 */
 	if (protoMajor == PROTO_MAJOR_V3)
 	{
+		if (kind != 'K')
+		{
+			pool_error("pool_do_auth: expect \"K\" got %c", kind);
+			return -1;
+		}
+
 		if ((length = pool_read_message_length(cp)) != 12)
 		{
 			pool_error("pool_do_auth: invalid messages length(%d) for BackendKeyData", length);
@@ -343,7 +347,6 @@ from pool_read_message_length and recheck the pg_hba.conf settings.");
 	 * OK, read pid and secret key
 	 */
 	sp = MASTER_CONNECTION(cp)->sp;
-	pid = -1;
 
 	for (i=0;i<NUM_BACKENDS;i++)
 	{
@@ -376,12 +379,14 @@ from pool_read_message_length and recheck the pg_hba.conf settings.");
 		}
 	}
 
-	if (pid == -1)
-	{
-		pool_error("pool_do_auth: all backends are down");
-		return -1;
-	}
-
+#ifdef NOT_USED
+	sp = MASTER_CONNECTION(cp)->sp;
+	cp->info->major = sp->major;
+	cp->info->minor = sp->minor;
+	strncpy(cp->info->database, sp->database, sizeof(cp->info->database) - 1);
+	strncpy(cp->info->user, sp->user, sizeof(cp->info->user) - 1);
+	cp->info->counter = 1;
+#endif
 	return pool_send_backend_key_data(frontend, pid, key, protoMajor);
 }
 
@@ -508,7 +513,7 @@ static POOL_STATUS pool_send_backend_key_data(POOL_CONNECTION *frontend, int pid
 }
 
 /*
- * perform clear text password authetication
+ * perform clear text password authentication
  */
 static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
@@ -521,7 +526,7 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 	/* master? */
 	if (IS_MASTER_NODE_ID(backend->db_node_id))
 	{
-		pool_write(frontend, "R", 1);	/* authenticaton */
+		pool_write(frontend, "R", 1);	/* authentication */
 		if (protoMajor == PROTO_MAJOR_V3)
 		{
 			len = htonl(8);
@@ -578,13 +583,13 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 	{
 		if ((ntohl(size) - 4) != backend->pwd_size)
 		{
-			pool_debug("do_clear_text_password; password size does not match in re-authetication");
+			pool_debug("do_clear_text_password; password size does not match in re-authentication");
 			return -1;
 		}
 
 		if (memcmp(password, backend->password, backend->pwd_size) != 0)
 		{
-			pool_debug("do_clear_text_password; password does not match in re-authetication");
+			pool_debug("do_clear_text_password; password does not match in re-authentication");
 			return -1;
 		}
 
@@ -660,7 +665,7 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 }
 
 /*
- * perform crypt authetication
+ * perform crypt authentication
  */
 static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
@@ -688,7 +693,7 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 	/* master? */
 	if (IS_MASTER_NODE_ID(backend->db_node_id))
 	{
-		pool_write(frontend, "R", 1);	/* authenticaton */
+		pool_write(frontend, "R", 1);	/* authentication */
 		if (protoMajor == PROTO_MAJOR_V3)
 		{
 			len = htonl(10);
@@ -827,7 +832,7 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 }
 
 /*
- * perform MD5 authetication
+ * perform MD5 authentication
  */
 static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
@@ -1007,7 +1012,7 @@ static int send_md5auth_request(POOL_CONNECTION *frontend, int protoMajor, char
 	int len;
 	int kind;
 
-	pool_write(frontend, "R", 1);	/* authenticaton */
+	pool_write(frontend, "R", 1);	/* authentication */
 	if (protoMajor == PROTO_MAJOR_V3)
 	{
 		len = htonl(12);
@@ -1042,17 +1047,17 @@ static int read_password_packet(POOL_CONNECTION *frontend, int protoMajor, 	char
 
 		if (pool_read(frontend, &k, sizeof(k)))
 		{
-			pool_debug("read_password_packet: failed to read password packet \"p\"");
+			pool_debug("read_password_packet_password: failed to read password packet \"p\"");
 			return -1;
 		}
 		if (k != 'p')
 		{
-			pool_error("read_password_packet: password packet does not start with \"p\"");
+			pool_error("read_password_packet_password: password packet does not start with \"p\"");
 			return -1;
 		}
 		if (pool_read(frontend, &size, sizeof(size)))
 		{
-			pool_error("read_password_packet: failed to read password packet size");
+			pool_error("read_password_packet_password: failed to read password packet size");
 			return -1;
 		}
 	}
@@ -1087,7 +1092,7 @@ static int read_password_packet(POOL_CONNECTION *frontend, int protoMajor, 	char
 /*
  * Send password packet to backend and receive authentication response
  * packet.  Return value is the last field of authentication
- * response. If it's 0, authentication was successfull.
+ * response. If it's 0, authentication was successful.
  * "password" must be null-terminated.
  */
 static int send_password_packet(POOL_CONNECTION *backend, int protoMajor, char *password)
@@ -1277,7 +1282,6 @@ signed char pool_read_kind(POOL_CONNECTION_POOL *cp)
 	char kind0, kind;
 	int i;
 
-	kind = -1;
 	kind0 = 0;
 
 	for (i=0;i<NUM_BACKENDS;i++)
@@ -1334,7 +1338,6 @@ int pool_read_int(POOL_CONNECTION_POOL *cp)
 	int data0, data;
 	int i;
 
-	data = -1;
 	data0 = 0;
 
 	for (i=0;i<NUM_BACKENDS;i++)
@@ -1374,51 +1377,13 @@ int pool_read_int(POOL_CONNECTION_POOL *cp)
  */
 void pool_random_salt(char *md5Salt)
 {
-	long rand = PostmasterRandom();
+	long rand = random();
 
 	md5Salt[0] = (rand % 255) + 1;
-	rand = PostmasterRandom();
+	rand = random();
 	md5Salt[1] = (rand % 255) + 1;
-	rand = PostmasterRandom();
+	rand = random();
 	md5Salt[2] = (rand % 255) + 1;
-	rand = PostmasterRandom();
+	rand = random();
 	md5Salt[3] = (rand % 255) + 1;
 }
-
-/*
- * PostmasterRandom
- */
-static long
-PostmasterRandom(void)
-{
-	extern struct timeval random_start_time;
-	static unsigned int random_seed = 0;
-
-	/*
-	 * Select a random seed at the time of first receiving a request.
-	 */
-	if (random_seed == 0)
-	{
-		do
-		{
-			struct timeval random_stop_time;
-
-			gettimeofday(&random_stop_time, NULL);
-
-			/*
-			 * We are not sure how much precision is in tv_usec, so we swap
-			 * the high and low 16 bits of 'random_stop_time' and XOR them
-			 * with 'random_start_time'. On the off chance that the result is
-			 * 0, we loop until it isn't.
-			 */
-			random_seed = random_start_time.tv_usec ^
-				((random_stop_time.tv_usec << 16) |
-				 ((random_stop_time.tv_usec >> 16) & 0xffff));
-		}
-		while (random_seed == 0);
-
-		srandom(random_seed);
-	}
-
-	return random();
-}
