@@ -5,7 +5,7 @@
 * pgpool: a language independent connection pool server for PostgreSQL
 * written by Tatsuo Ishii
 *
-* Copyright (c) 2003-2015	PgPool Global Development Group
+* Copyright (c) 2003-2013	PgPool Global Development Group
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
@@ -79,7 +79,6 @@ POOL_CONNECTION *pool_open(int fd)
 	if (cp->hp == NULL)
 	{
 		pool_error("pool_open: malloc failed");
-		free(cp);
 		return NULL;
 	}
 	cp->bufsz = READBUFSZ;
@@ -89,8 +88,6 @@ POOL_CONNECTION *pool_open(int fd)
 	cp->sbufsz = 0;
 	cp->buf2 = NULL;
 	cp->bufsz2 = 0;
-	cp->buf3 = NULL;
-	cp->bufsz3 = 0;
 
 	cp->fd = fd;
 	return cp;
@@ -114,8 +111,6 @@ void pool_close(POOL_CONNECTION *cp)
 		free(cp->sbuf);
 	if (cp->buf2)
 		free(cp->buf2);
-	if (cp->buf3)
-		free(cp->buf3);
 	pool_discard_params(&cp->params);
 
 	pool_ssl_close(cp);
@@ -173,7 +168,7 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 
 			if (cp->isbackend)
 			{
-				/* if fail_over_on_backend_erro is true, then trigger failover */
+				/* if fail_over_on_backend_error is true, then trigger failover */
 				if (pool_config->fail_over_on_backend_error)
 				{
 					notice_backend_error(cp->db_node_id);
@@ -298,7 +293,7 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 
 			if (cp->isbackend)
 			{
-				/* if fail_over_on_backend_erro is true, then trigger failover */
+				/* if fail_over_on_backend_error is true, then trigger failover */
 				if (pool_config->fail_over_on_backend_error)
 				{
 					notice_backend_error(cp->db_node_id);
@@ -439,7 +434,7 @@ int pool_flush_it(POOL_CONNECTION *cp)
 			}
 			else if (FD_ISSET(cp->fd, &exceptmask))
 			{
-				pool_log("pool_flush_it: exception occured");
+				pool_log("pool_flush_it: exception occurred");
 				cp->wbufpo = 0;
 				return -1;
 			}
@@ -512,7 +507,7 @@ int pool_flush(POOL_CONNECTION *cp)
 	{
 		if (cp->isbackend)
 		{
-			/* if fail_over_on_backend_erro is true, then trigger failover */
+			/* if fail_over_on_backend_error is true, then trigger failover */
 			if (pool_config->fail_over_on_backend_error)
 			{
 				notice_backend_error(cp->db_node_id);
@@ -719,16 +714,6 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 	return cp->sbuf;
 }
 
-/*
- * Set db node id to connection.
- */
-void pool_set_db_node_id(POOL_CONNECTION *con, int db_node_id)
-{
-	if (!con)
-		return;
-	con->db_node_id = db_node_id;
-}
-
 /*
  * returns the byte length of str, including \0, no more than upper.
  * if encountered \0, flag is set to non 0.
@@ -864,7 +849,6 @@ int pool_unread(POOL_CONNECTION *cp, void *data, int len)
 			return -1;
 		}
 		cp->hp = p;
-		cp->bufsz = realloc_size;
 	}
 	if (cp->len != 0)
 		memmove(p + len, cp->hp + cp->po, cp->len);
@@ -894,8 +878,8 @@ int pool_push(POOL_CONNECTION *cp, void *data, int len)
 	}
 	else
 	{
-		cp->buf3 = realloc(cp->buf3, cp->bufsz3 + len);
 		p = cp->buf3 + cp->bufsz3;
+		cp->buf3 = realloc(cp->buf3, cp->bufsz3 + len);
 	}
 
 	memcpy(p, data, len);
